# STB 34.101.45-2013
# Copyright (C) 2015  NeverWalkAloner

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see  <http://www.gnu.org/licenses/>.
import random
from publickey.ec import ECPoint
import symmetric.belt as belt


class STB:
    # Initialization of STB object
    # p, a, b, q, g_y - int, EC characteristics
    # l - int, security bit level
    def __init__(self, p, a, b, q, g_y, l):
        self.g_point = ECPoint(0, g_y, a, b, p)
        self.q = q
        self.a = a
        self.b = b
        self.p = p
        self.l = l

    # generate key pair
    def gen_keys(self):
        d = random.randint(1, self.q - 1)
        q_point = d * self.g_point
        return d, q_point

    # sign message with private key
    # message, private_key - int
    def sign(self, message, private_key, k=0):
        if k == 0:
            k = random.randint(1, self.q-1)
        oid = 0x06092A7000020022651F51
        r_point = k * self.g_point
        hash_value = self.belt_hash(message)
        value_to_hash = int2list(oid) + int2list(reverse(r_point.x)) + int2list(hash_value)
        s0 = self.belt_hash(list2int(value_to_hash))
        s0 = int2list(s0)[:16]
        s0 = list2int(s0)
        s1 = (k - reverse(hash_value) - (reverse(s0) + 2 ** self.l) * private_key) % self.q
        return s0, reverse(s1)

    # verify signatute
    # message, sign - int
    # pub_key - tuple
    def verify(self, message, pub_key, sign):
        if sign[1] > self.q:
            return False
        public_key = ECPoint(pub_key[0], pub_key[1], self.a, self.b, self.p)
        sign_tmp = reverse(sign[0]), reverse(sign[1])
        oid = 0x06092A7000020022651F51
        hash_value = self.belt_hash(message)
        r_point = ((sign_tmp[1] + reverse(hash_value)) % self.q) * self.g_point + \
                  (sign_tmp[0] + 2 ** self.l) * public_key
        if r_point == float('inf'):
            return False
        value_to_hash = int2list(oid) + int2list(reverse(r_point.x)) + int2list(hash_value)
        t = self.belt_hash(list2int(value_to_hash))
        t = int2list(t)[:16]
        if list2int(t) != sign[0]:
            return False
        return True

    # compute belt hash of message
    # message - int
    def belt_hash(self, message):
        h = 0xB194BAC80A08F53B366D008E584A5DE48504FA9D1BB6C7AC252E72C202FDCE0D
        h = int2list(h)
        bit_l = message.bit_length()
        l_message = int2list(message)
        len_dif = 8 - l_message[0].bit_length()
        bit_l += len_dif
        s = [0 for i in range(16)]
        if len(l_message) % 32 != 0:
            l_message = l_message + [0] * (32 - (len(l_message)) % 32)
        for i in range(len(l_message) // 32):
            m_list = l_message[i*32:(i+1)*32]
            m_list = m_list + h
            s = xor(s, self.sigma1(m_list))
            h = self.sigma2(m_list)
        bit_l = int2list(bit_l)
        if len(bit_l) < 16:
            bit_l = list(reversed(bit_l)) + [0] * (16 - len(bit_l))
        return list2int(self.sigma2(bit_l + s + h))

    # belt hash supporting function
    def sigma1(self, m_list):
        my_belt = belt.belt(m_list[:32])
        m = xor(m_list[32:48], m_list[48:64])
        sigma1 = xor(my_belt.encryption(m), m_list[32:48])
        sigma1 = xor(sigma1, m_list[48:64])
        return sigma1

    # belt hash supporting function
    def sigma2(self, m_list):
        s = self.sigma1(m_list)
        ones = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff]
        teta1 = s + m_list[48:64]
        teta2 = xor(s, ones) + m_list[32:48]
        u1 = m_list[0:16]
        u2 = m_list[16:32]
        my_belt = belt.belt(teta1)
        tmp = xor(my_belt.encryption(u1), u1)
        my_belt2 = belt.belt(teta2)
        tm2 = xor(my_belt2.encryption(u2), u2)
        return tmp + tm2


# Represent number as list of bytes
def int2list(x):
    l = x.bit_length()
    if l % 8 == 0:
        l -= 8
    else:
        l -= (l % 8)
    l = list(range(l, -1, -8))
    return [x >> i & 0xff for i in l]


# Represent list of bytes as number
def list2int(x):
    l = [8 * i for i in range(len(x)-1, -1, -1)]
    return sum([x[i] << l[i] for i in range(len(x))])


# Reverse bytes of number x
def reverse(x):
    l = int2list(x)
    l.reverse()
    return list2int(l)


# XOR x and y byte array
def xor(x, y):
    return [x[i] ^ y[i] for i in range(len(x))]



# STB 34.101.31-2011 with 128-bit block and 256-bit key
# Copyright (C) 2015  NeverWalkAloner

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class belt:
    def __init__(self, key):
        self.H = [0xB1, 0x94, 0xBA, 0xC8, 0x0A, 0x08, 0xF5, 0x3B, 0x36, 0x6D, 0x00, 0x8E, 0x58, 0x4A, 0x5D, 0xE4,
                  0x85, 0x04, 0xFA, 0x9D, 0x1B, 0xB6, 0xC7, 0xAC, 0x25, 0x2E, 0x72, 0xC2, 0x02, 0xFD, 0xCE, 0x0D,
                  0x5B, 0xE3, 0xD6, 0x12, 0x17, 0xB9, 0x61, 0x81, 0xFE, 0x67, 0x86, 0xAD, 0x71, 0x6B, 0x89, 0x0B,
                  0x5C, 0xB0, 0xC0, 0xFF, 0x33, 0xC3, 0x56, 0xB8, 0x35, 0xC4, 0x05, 0xAE, 0xD8, 0xE0, 0x7F, 0x99,
                  0xE1, 0x2B, 0xDC, 0x1A, 0xE2, 0x82, 0x57, 0xEC, 0x70, 0x3F, 0xCC, 0xF0, 0x95, 0xEE, 0x8D, 0xF1,
                  0xC1, 0xAB, 0x76, 0x38, 0x9F, 0xE6, 0x78, 0xCA, 0xF7, 0xC6, 0xF8, 0x60, 0xD5, 0xBB, 0x9C, 0x4F,
                  0xF3, 0x3C, 0x65, 0x7B, 0x63, 0x7C, 0x30, 0x6A, 0xDD, 0x4E, 0xA7, 0x79, 0x9E, 0xB2, 0x3D, 0x31,
                  0x3E, 0x98, 0xB5, 0x6E, 0x27, 0xD3, 0xBC, 0xCF, 0x59, 0x1E, 0x18, 0x1F, 0x4C, 0x5A, 0xB7, 0x93,
                  0xE9, 0xDE, 0xE7, 0x2C, 0x8F, 0x0C, 0x0F, 0xA6, 0x2D, 0xDB, 0x49, 0xF4, 0x6F, 0x73, 0x96, 0x47,
                  0x06, 0x07, 0x53, 0x16, 0xED, 0x24, 0x7A, 0x37, 0x39, 0xCB, 0xA3, 0x83, 0x03, 0xA9, 0x8B, 0xF6,
                  0x92, 0xBD, 0x9B, 0x1C, 0xE5, 0xD1, 0x41, 0x01, 0x54, 0x45, 0xFB, 0xC9, 0x5E, 0x4D, 0x0E, 0xF2,
                  0x68, 0x20, 0x80, 0xAA, 0x22, 0x7D, 0x64, 0x2F, 0x26, 0x87, 0xF9, 0x34, 0x90, 0x40, 0x55, 0x11,
                  0xBE, 0x32, 0x97, 0x13, 0x43, 0xFC, 0x9A, 0x48, 0xA0, 0x2A, 0x88, 0x5F, 0x19, 0x4B, 0x09, 0xA1,
                  0x7E, 0xCD, 0xA4, 0xD0, 0x15, 0x44, 0xAF, 0x8C, 0xA5, 0x84, 0x50, 0xBF, 0x66, 0xD2, 0xE8, 0x8A,
                  0xA2, 0xD7, 0x46, 0x52, 0x42, 0xA8, 0xDF, 0xB3, 0x69, 0x74, 0xC5, 0x51, 0xEB, 0x23, 0x29, 0x21,
                  0xD4, 0xEF, 0xD9, 0xB4, 0x3A, 0x62, 0x28, 0x75, 0x91, 0x14, 0x10, 0xEA, 0x77, 0x6C, 0xDA, 0x1D]
        key = [self.list2int(key[i:i+4]) for i in range(0, len(key), 4)]
        self.k = [key[i % 8] for i in range(56)]

    # RotHi operation
    def circularleftshift(self, value, k):
        bitlength = 32
        return (value << (k % bitlength) & 2 ** bitlength - 1) ^ (value >> (bitlength - k) % bitlength)

    # Represent 32-bit number as list of bytes
    def int2list(self, x):
        return [x >> i & 0xff for i in [24, 16, 8, 0]]

    # Represent list of bytes as 32-bit number
    def list2int(self, x):
        l = [24, 16, 8, 0]
        return sum([x[i] << l[i] for i in range(4)])

    # Modular Substraction
    def modsub(self, x, y):
        mod = 2**32
        return (x - y) % mod

    # Modular addition
    def modadd(self, *x):
        mod = 2**32
        res = 0
        for el in x:
            res = (res + self.reverse(el)) % mod
        return self.reverse(res)

    # H-transformation: replacing byte with another value from table
    def htransformation(self, x):
        return self.H[x]

    # G-transformation
    def gtransformation(self, x, k):
        res = self.list2int([self.htransformation(i) for i in self.int2list(x)])
        return self.reverse(self.circularleftshift(self.reverse(res), k))

    # Reverse bytes in word x
    def reverse(self, x):
        l = self.int2list(x)
        l.reverse()
        return self.list2int(l)

    # Encrypt input m represented as list of bytes using key represented as list of bytes
    def encryption(self, m):
        a, b, c, d = [self.list2int(m[i:i+4]) for i in range(0, len(m), 4)]
        for i in range(8):
            b = b ^ self.gtransformation(self.modadd(a, self.k[7*i+0]), 5)
            c = c ^ self.gtransformation(self.modadd(d, self.k[7*i+1]), 21)
            a = self.reverse(self.modsub(self.reverse(a),
                                         self.reverse(self.gtransformation(self.modadd(b, self.k[7*i+2]), 13))))
            e = (self.gtransformation(self.modadd(b, c, self.k[7*i+3]), 21)) ^ self.reverse(i+1)
            b = self.modadd(b, e)
            c = self.reverse(self.modsub(self.reverse(c), self.reverse(e)))
            d = self.modadd(d, self.gtransformation(self.modadd(c, self.k[7*i+4]), 13))
            b = b ^ self.gtransformation(self.modadd(a, self.k[7*i+5]), 21)
            c = c ^ self.gtransformation(self.modadd(d, self.k[7*i+6]), 5)
            a, b = b, a
            c, d = d, c
            b, c = c, b
        a = self.int2list(a)
        b = self.int2list(b)
        c = self.int2list(c)
        d = self.int2list(d)
        return b + d + a + c

    # Decrypt input m represented as list of bytes using key represented as list of bytes
    def decryption(self, m):
        a, b, c, d = [self.list2int(m[i:i+4]) for i in range(0, len(m), 4)]
        for i in reversed(range(8)):
            b = b ^ self.gtransformation(self.modadd(a, self.k[7*i+6]), 5)
            c = c ^ self.gtransformation(self.modadd(d, self.k[7*i+5]), 21)
            a = self.reverse(self.modsub(self.reverse(a),
                                         self.reverse(self.gtransformation(self.modadd(b, self.k[7*i+4]), 13))))
            e = (self.gtransformation(self.modadd(b, c, self.k[7*i+3]), 21)) ^ self.reverse(i+1)
            b = self.modadd(b, e)
            c = self.reverse(self.modsub(self.reverse(c), self.reverse(e)))
            d = self.modadd(d, self.gtransformation(self.modadd(c, self.k[7*i+2]), 13))
            b = b ^ self.gtransformation(self.modadd(a, self.k[7*i+1]), 21)
            c = c ^ self.gtransformation(self.modadd(d, self.k[7*i+0]), 5)
            a, b = b, a
            c, d = d, c
            a, d = d, a
        a = self.int2list(a)
        b = self.int2list(b)
        c = self.int2list(c)
        d = self.int2list(d)
        return c + a + d + b



# GOST 34.10?2012
# Copyright (C) 2015  NeverWalkAloner

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see  <http://www.gnu.org/licenses/>.
import random
from publickey.ec import ECPoint


class DSGOST:
    # p - int, EC module
    # a, b - int, EC coefficients
    # q - int, order of point P
    # p_x, p_y - int, point P coordinates
    def __init__(self, p, a, b, q, p_x, p_y):
        self.p_point = ECPoint(p_x, p_y, a, b, p)
        self.q = q
        self.a = a
        self.b = b
        self.p = p

    # generate key pair
    def gen_keys(self):
        d = random.randint(1, self.q - 1)
        q_point = d * self.p_point
        return d, q_point

    # sing message
    # message - int
    # private_key - int
    def sign(self, message, private_key, k=0):
        e = message % self.q
        if e == 0:
            e = 1
        if k == 0:
            k = random.randint(1, self.q - 1)
        r, s = 0, 0
        while r == 0 or s == 0:
            c_point = k * self.p_point
            r = c_point.x % self.q
            s = (r * private_key + k * e) % self.q
        return r, s

    # verify singed message
    # message - int
    # sign - tuple
    # public_key - ECPoint
    def verify(self, message, sign, public_key):
        e = message % self.q
        if e == 0:
            e = 1
        nu = ECPoint._mod_inverse(e, self.q)
        z1 = (sign[1] * nu) % self.q
        z2 = (-sign[0] * nu) % self.q
        c_point = z1 * self.p_point + z2 * public_key
        r = c_point.x % self.q
        if r == sign[0]:
            return True
        return False
